/*
 * Programmieren fuer Physiker, SS 2023, Blatt 10, Aufgabe 26
 *
 * Normalverteilte Zufallszahlen mit Polar-Transformation
 */

#include <iostream>
#include <cstdlib>
#include <ctime>
#include <cmath>
using namespace std ;

class normal_generator 
{
public:
    normal_generator() ;
    double get() ;
private:
    double saved_number ;
    bool have_one_more ;
    double box_random() ;
} ;

normal_generator::normal_generator()
{
    srand( (unsigned int)time(0)) ;
    have_one_more = false ;
}

// berechne gleichverteilte Zufallszahl in -1..1
double normal_generator::box_random() 
{
    return 2.0*rand()/double(RAND_MAX)-1.0 ;
}

double normal_generator::get() 
{
    if (have_one_more)
    {
	have_one_more = false ;
	return saved_number ;
    }
    
    double v1, v2, s ;
    do
    {
	v1 = box_random() ;
	v2 = box_random() ;
	s = v1*v1 + v2*v2 ;
    }
    while (s>=1 or s==0) ;
    double fac = sqrt(-2*log(s)/s) ;
    
    have_one_more = true ;
    saved_number = v2 * fac ;
    return v1 * fac ;
}

int main()
{
    cout << "Normal-Verteilung" << endl ;

    normal_generator ng ;

    // 1. Schritt: Mittelwert und Varianz
    double m=0.0, sigma=0.0 ;
    const long Nexp = 100000 ;
    cout << "Anzahl der Experimente: " << Nexp << endl ;

    for( long i=0; i<Nexp; ++i)
    {
	double x = ng.get() ;
	m += x ;
	sigma += x*x ;
    }

    m /= double(Nexp) ;
    sigma = sqrt(sigma/double(Nexp) - m*m) ;

    cout << "Mittelwert: " << m << endl ;
    cout << "Standardabweichung: " << sigma << endl ;
    
    // 2. Schritt: Abweichungen ein-, zwei- drei-sigma
    const int nsigma = 3 ;
    long a[nsigma] = {0,0,0} ; 
    for( long i=0; i<Nexp; ++i)
    {
	double x = ng.get() ;
	for( int j=0; j<nsigma; ++j)
	    if (abs(x-m)>(j+1)*sigma) 
		++a[j] ;
    }

    cout << "Wie oft kommt die Abweichung von ... "<< endl ;
    for( int j=0; j<nsigma; ++j)
	cout << " .... " << j+1 << " mal sigma vor: " 
	     << double(a[j])/double(Nexp)*100 << " % " << endl ;
}
	

