/*
 * Programmieren fuer Physiker, SS 2023, Blatt 10, Aufgabe 27
 *
 * Das N-Damen Problem
 *
 * Es sind N Damen auf einem N x N Schachbrett so anzuordnen,
 * dass keine Dame eine andere schlagen kann.
 * Dieses Programm findet saemtliche Anordnungen durch Ausprobieren
 * und gibt die Anzahl der Loesungen aus.
 *
 * Loesung mit Klasse zur Verwaltung des Spielfelds
 */

#include <iostream>
#include <string>
#include <cstdlib>

using namespace std ;

const int BrettGroesse = 8 ;

// ------------------------------ Start Class Spielfeld

class Spielfeld 
{
private:
    int AnzDamen ;
    int DameInSpalte[BrettGroesse+1] ; // Indexbereich 1..BrettGroesse
    // Alle Koordinaten hier 1..Brettgroesse 
public:
    Spielfeld() { AnzDamen=0 ; } // Konstruktor
    void SetzeDameInSpalte( int) ;
    void EntferneLetzteDame() ;
    bool IstSchlagbar( int, int) ;
    void BrettAusgabe() ;
    int GetAnzDamen() { return AnzDamen; } ;
} ;

void Spielfeld::SetzeDameInSpalte( int Spalte) 
{
    if ((Spalte<1) or (Spalte>BrettGroesse) or (AnzDamen>=BrettGroesse))
    {
	cout << "FEHLER: Dame ausserhalb des Bretts platziert." << endl ;
	exit(1) ;
    }
    ++AnzDamen ;
    DameInSpalte[AnzDamen] = Spalte ;
}

void Spielfeld::EntferneLetzteDame() 
{
   if (AnzDamen<=0)
    {
	cout << "FEHLER: Keine Dame zum Entfernen vorhanden." << endl ;
	exit(1) ;
    }
    --AnzDamen ;
}

bool Spielfeld::IstSchlagbar( int TestZl, int TestSp) 
// ueberprueft ob das Feld (TestZl, TestSp) von einer
// bereits vorhandenen Dame direkt erreichbar ist
{ 
    for( int zl=1; zl<=AnzDamen; ++zl) 
    {
	// Steht eine bereits platzierte Dame in der Kandidaten-Zeile?
	if (TestZl==zl) return true ;
	// Steht eine bereits platzierte Dame in der Kandidaten-Spalte?
	if( TestSp==DameInSpalte[zl]) return true ;
	// Ist eine bereit platzierte Dame diagonal schlagbar?
	int DiffSp=TestSp-DameInSpalte[zl] ;
	int DiffZl=TestZl-zl ;
	if (abs(DiffSp)==abs(DiffZl)) return true ;
    } 
    return false ;
}

void Spielfeld::BrettAusgabe() 
// Kleine ASCII Ausgabe des Spielbretts
{
    string FeldString ;
    for( int zl=1; zl<=BrettGroesse; ++zl)
    {
	for( int sp=1; sp<=BrettGroesse; ++sp)
	{
	    if(( sp+zl)%2 == 0) FeldString="  " ;
	    else FeldString =".." ;
	    if( zl<=AnzDamen)
		if( sp==DameInSpalte[zl]) FeldString="DD" ;
	    cout << FeldString ;
	}
	cout << endl ;
    }
}
	    
// ------------------------------- Ende Class Spielfeld

int PlatziereDamen( Spielfeld SpielSituation) 
// Platziere Damen gemaess "Unschlagbar-Forderung". Vorgehen:
// Probiere saemtliche Postionen in der aktuellen Spalte aus
// und teste bei erfolgreichen Konstellationen (mittels Rekursion)
// die naechste Spalte.
// Rueckgabe: Anzahl der Loesungen
//       (jeweils im Suchbaum unterhalb der aktuellen Konstellation)
{  
    int AnzLoes = 0 ;
    int TestZeile = SpielSituation.GetAnzDamen()+1 ;
    for( int TestSp=1; TestSp<=BrettGroesse; ++TestSp)
    {
	if( not SpielSituation.IstSchlagbar(TestZeile, TestSp))
	{
	    SpielSituation.SetzeDameInSpalte( TestSp) ;
	    // cout << "Aktuelle Konfig:" << endl ;
	    // SpielSituation.BrettAusgabe() ;
	    if( TestZeile==BrettGroesse) 
	    { 
		cout << "Loesung gefunden" << endl ;
		SpielSituation.BrettAusgabe() ;
		cout << endl ;
		++AnzLoes ;
	    }
	    else
		AnzLoes += PlatziereDamen( SpielSituation) ;
	    SpielSituation.EntferneLetzteDame() ;
	}
    }
    return AnzLoes ;
}

int main( void)
{
    int AnzahlLoesungen ;
    Spielfeld DamenKonfig ;
    AnzahlLoesungen = PlatziereDamen( DamenKonfig) ;
    cout << "Gesamtzahl an Loesungen: " << AnzahlLoesungen << endl ;
}


